#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
# from distutils.extension import Extension
import os

from setuptools import dist

dist.Distribution().fetch_build_eggs(['Cython>=0.29.32', 'numpy==1.24.0'])

import numpy as np

try:
    from Cython.Distutils import build_ext
    from Cython.Build import cythonize
except ImportError:
    # create closure for deferred import
    def cythonize(*args, **kwargs):
        from Cython.Build import cythonize
        return cythonize(*args, **kwargs)


    def build_ext(*args, **kwargs):
        from Cython.Distutils import build_ext
        return build_ext(*args, **kwargs)

from setuptools import find_packages
from setuptools import setup, Extension

from octobot_trading import PROJECT_NAME, VERSION

PACKAGES = find_packages(exclude=["tests"])

packages_list = [
    "octobot_trading.octobot_channel_consumer",
    "octobot_trading.util.initializable",
    "octobot_trading.util.config_util",
    "octobot_trading.exchange_data.exchange_symbols_data",
    "octobot_trading.exchange_data.exchange_symbol_data",
    "octobot_trading.exchange_data.ticker.ticker_manager",
    "octobot_trading.exchange_data.ticker.channel.ticker",
    "octobot_trading.exchange_data.ticker.channel.ticker_updater",
    "octobot_trading.exchange_data.ticker.channel.ticker_updater_simulator",
    "octobot_trading.exchange_data.contracts.margin_contract",
    "octobot_trading.exchange_data.contracts.future_contract",
    "octobot_trading.exchange_data.order_book.order_book_manager",
    "octobot_trading.exchange_data.order_book.channel.order_book",
    "octobot_trading.exchange_data.order_book.channel.order_book_updater_simulator",
    "octobot_trading.exchange_data.order_book.channel.order_book_updater",
    "octobot_trading.exchange_data.funding.funding_manager",
    "octobot_trading.exchange_data.funding.channel.funding",
    "octobot_trading.exchange_data.funding.channel.funding_updater",
    "octobot_trading.exchange_data.funding.channel.funding_updater_simulator",
    "octobot_trading.exchange_data.kline.kline_manager",
    "octobot_trading.exchange_data.kline.channel.kline",
    "octobot_trading.exchange_data.kline.channel.kline_updater",
    "octobot_trading.exchange_data.kline.channel.kline_updater_simulator",
    "octobot_trading.exchange_data.recent_trades.recent_trades_manager",
    "octobot_trading.exchange_data.recent_trades.channel.recent_trade",
    "octobot_trading.exchange_data.recent_trades.channel.recent_trade_updater_simulator",
    "octobot_trading.exchange_data.recent_trades.channel.recent_trade_updater",
    "octobot_trading.exchange_data.prices.price_events_manager",
    "octobot_trading.exchange_data.prices.prices_manager",
    "octobot_trading.exchange_data.prices.channel.price",
    "octobot_trading.exchange_data.prices.channel.prices_updater_simulator",
    "octobot_trading.exchange_data.prices.channel.prices_updater",
    "octobot_trading.exchange_data.ohlcv.candles_manager",
    "octobot_trading.exchange_data.ohlcv.preloaded_candles_manager",
    "octobot_trading.exchange_data.ohlcv.candles_adapter",
    "octobot_trading.exchange_data.ohlcv.channel.ohlcv_updater",
    "octobot_trading.exchange_data.ohlcv.channel.ohlcv_updater_simulator",
    "octobot_trading.exchange_data.ohlcv.channel.ohlcv",
    "octobot_trading.personal_data.exchange_personal_data",
    "octobot_trading.personal_data.state",
    "octobot_trading.personal_data.trades.trades_manager",
    "octobot_trading.personal_data.trades.trade",
    "octobot_trading.personal_data.trades.trade_factory",
    "octobot_trading.personal_data.trades.channel.trades_updater",
    "octobot_trading.personal_data.trades.channel.trades",
    "octobot_trading.personal_data.orders.order",
    "octobot_trading.personal_data.orders.order_adapter",
    "octobot_trading.personal_data.orders.decimal_order_adapter",
    "octobot_trading.personal_data.orders.orders_manager",
    "octobot_trading.personal_data.orders.order_state",
    "octobot_trading.personal_data.orders.order_group",
    "octobot_trading.personal_data.orders.order_util",
    "octobot_trading.personal_data.orders.order_factory",
    "octobot_trading.personal_data.orders.groups.balanced_take_profit_and_stop_order_group",
    "octobot_trading.personal_data.orders.groups.one_cancels_the_other_order_group",
    "octobot_trading.personal_data.orders.states.pending_creation_order_state",
    "octobot_trading.personal_data.orders.states.pending_creation_chained_order_state",
    "octobot_trading.personal_data.orders.states.fill_order_state",
    "octobot_trading.personal_data.orders.states.cancel_order_state",
    "octobot_trading.personal_data.orders.states.close_order_state",
    "octobot_trading.personal_data.orders.states.order_state_factory",
    "octobot_trading.personal_data.orders.states.open_order_state",
    "octobot_trading.personal_data.orders.types.unknown_order",
    "octobot_trading.personal_data.orders.types.unsupported_order",
    "octobot_trading.personal_data.orders.types.market.buy_market_order",
    "octobot_trading.personal_data.orders.types.market.market_order",
    "octobot_trading.personal_data.orders.types.market.sell_market_order",
    "octobot_trading.personal_data.orders.types.trailing.trailing_stop_limit_order",
    "octobot_trading.personal_data.orders.types.trailing.trailing_stop_order",
    "octobot_trading.personal_data.orders.types.limit.take_profit_limit_order",
    "octobot_trading.personal_data.orders.types.limit.stop_loss_limit_order",
    "octobot_trading.personal_data.orders.types.limit.limit_order",
    "octobot_trading.personal_data.orders.types.limit.sell_limit_order",
    "octobot_trading.personal_data.orders.types.limit.stop_loss_order",
    "octobot_trading.personal_data.orders.types.limit.buy_limit_order",
    "octobot_trading.personal_data.orders.types.limit.take_profit_order",
    "octobot_trading.personal_data.orders.channel.orders_updater_simulator",
    "octobot_trading.personal_data.orders.channel.orders",
    "octobot_trading.personal_data.orders.channel.orders_updater",
    "octobot_trading.personal_data.portfolios.portfolio_value_holder",
    "octobot_trading.personal_data.portfolios.portfolio_manager",
    "octobot_trading.personal_data.portfolios.sub_portfolio",
    "octobot_trading.personal_data.portfolios.portfolio",
    "octobot_trading.personal_data.portfolios.asset",
    "octobot_trading.personal_data.portfolios.portfolio_factory",
    "octobot_trading.personal_data.portfolios.portfolio_profitability",
    "octobot_trading.personal_data.portfolios.portfolio_util",
    "octobot_trading.personal_data.portfolios.assets.future_asset",
    "octobot_trading.personal_data.portfolios.assets.margin_asset",
    "octobot_trading.personal_data.portfolios.assets.spot_asset",
    "octobot_trading.personal_data.portfolios.types.spot_portfolio",
    "octobot_trading.personal_data.portfolios.types.future_portfolio",
    "octobot_trading.personal_data.portfolios.types.margin_portfolio",
    "octobot_trading.personal_data.portfolios.channel.balance_updater",
    "octobot_trading.personal_data.portfolios.channel.balance_updater_simulator",
    "octobot_trading.personal_data.portfolios.channel.balance",
    "octobot_trading.personal_data.portfolios.history.historical_asset_value",
    "octobot_trading.personal_data.portfolios.history.historical_asset_value_factory",
    "octobot_trading.personal_data.portfolios.history.historical_portfolio_value_manager",
    "octobot_trading.personal_data.positions.position",
    "octobot_trading.personal_data.positions.position_factory",
    "octobot_trading.personal_data.positions.position_state",
    "octobot_trading.personal_data.positions.position_util",
    "octobot_trading.personal_data.positions.positions_manager",
    "octobot_trading.personal_data.positions.channel.positions",
    "octobot_trading.personal_data.positions.channel.positions_updater",
    "octobot_trading.personal_data.positions.channel.positions_updater_simulator",
    "octobot_trading.personal_data.positions.types.inverse_position",
    "octobot_trading.personal_data.positions.types.linear_position",
    "octobot_trading.personal_data.positions.states.liquidate_position_state",
    "octobot_trading.personal_data.positions.states.idle_position_state",
    "octobot_trading.personal_data.positions.states.active_position_state",
    "octobot_trading.personal_data.positions.states.position_state_factory",
    "octobot_trading.personal_data.transactions.transaction",
    "octobot_trading.personal_data.transactions.transaction_factory",
    "octobot_trading.personal_data.transactions.transactions_manager",
    "octobot_trading.personal_data.transactions.types.blockchain_transaction",
    "octobot_trading.personal_data.transactions.types.fee_transaction",
    "octobot_trading.personal_data.transactions.types.realised_pnl_transaction",
    "octobot_trading.personal_data.transactions.types.transfer_transaction",
    "octobot_trading.modes.modes_factory",
    "octobot_trading.modes.modes_util",
    "octobot_trading.modes.channel.abstract_mode_producer",
    "octobot_trading.modes.channel.mode",
    "octobot_trading.modes.channel.abstract_mode_consumer",
    "octobot_trading.exchanges.exchanges",
    "octobot_trading.exchanges.exchange_manager",
    "octobot_trading.exchanges.abstract_exchange",
    "octobot_trading.exchanges.abstract_websocket_exchange",
    "octobot_trading.exchanges.basic_exchange_wrapper",
    "octobot_trading.exchanges.exchange_websocket_factory",
    "octobot_trading.exchanges.config.exchange_config_data",
    "octobot_trading.exchanges.config.backtesting_exchange_config",
    "octobot_trading.exchanges.exchange_builder",
    "octobot_trading.exchanges.exchange_channels",
    "octobot_trading.exchanges.exchange_factory",
    "octobot_trading.exchanges.connectors.ccxt.ccxt_client_util",
    "octobot_trading.exchanges.connectors.ccxt.ccxt_connector",
    "octobot_trading.exchanges.connectors.ccxt.ccxt_websocket_connector",
    "octobot_trading.exchanges.connectors.simulator.exchange_simulator_connector",
    "octobot_trading.exchanges.traders.trader",
    "octobot_trading.exchanges.traders.trader_simulator",
    "octobot_trading.exchanges.util.exchange_market_status_fixer",
    "octobot_trading.exchanges.util.websockets_util",
    "octobot_trading.exchanges.util.exchange_util",
    "octobot_trading.exchanges.types.rest_exchange",
    "octobot_trading.exchanges.types.websocket_exchange",
    "octobot_trading.exchanges.implementations.default_websocket_exchange",
    "octobot_trading.exchanges.implementations.default_rest_exchange",
    "octobot_trading.exchanges.implementations.exchange_simulator",
    "octobot_trading.exchange_channel",
]

ext_modules = [
    Extension(package, [f"{package.replace('.', '/')}.py"], include_dirs=[np.get_include()])
    for package in packages_list]

# long description from README file
with open('README.md', encoding='utf-8') as f:
    DESCRIPTION = f.read()

REQUIRED = open('requirements.txt').readlines()
REQUIRES_PYTHON = '>=3.8'
CYTHON_DEBUG = False if not os.getenv('CYTHON_DEBUG') else os.getenv('CYTHON_DEBUG')

if CYTHON_DEBUG:
    from Cython.Compiler.Options import get_directive_defaults

    get_directive_defaults()['cache_builtins'] = False

setup(
    name=PROJECT_NAME,
    version=VERSION,
    url='https://github.com/Drakkar-Software/OctoBot-Trading',
    license='LGPL-3.0',
    author='Drakkar-Software',
    author_email='drakkar-software@protonmail.com',
    description='OctoBot project trading package',
    packages=PACKAGES,
    include_package_data=True,
    long_description=DESCRIPTION,
    include_dirs=[np.get_include()],
    cmdclass={'build_ext': build_ext},
    tests_require=["pytest"],
    test_suite="tests",
    zip_safe=False,
    data_files=[],
    setup_requires=REQUIRED,
    install_requires=REQUIRED,
    ext_modules=cythonize(ext_modules, gdb_debug=CYTHON_DEBUG),
    python_requires=REQUIRES_PYTHON,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Cython',
        'Operating System :: OS Independent',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX'
    ],
)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.constants as constants

from tests_additional.real_exchanges import get_exchange_manager
from tests_additional.real_exchanges.real_exchange_tester import RealExchangeTester


class RealFuturesExchangeTester(RealExchangeTester):
    EXCHANGE_TYPE = enums.ExchangeTypes.FUTURE.value

    async def test_get_funding_rate(self):
        pass

    async def get_funding_rate(self, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME, self.get_config()) as exchange_manager:
            return (
                await exchange_manager.exchange.get_funding_rate(self.SYMBOL, **kwargs),
                exchange_manager.exchange.parse_funding(
                    await exchange_manager.exchange.get_price_ticker(self.SYMBOL),
                    from_ticker=True
                )
            )

    def _check_funding_rate(
        self,
        funding_rate,
        has_rate=True,
        has_last_time=True,
        has_next_rate=True,
        has_next_time=True
    ):
        """
        Used data are
        - funding rate (value)
        - last_updated (timestamp)
        - predicted_funding_rate (value)
        - next_update (timestamp)
        """
        assert funding_rate
        if has_rate:
            assert funding_rate[enums.ExchangeConstantsFundingColumns.FUNDING_RATE.value]
        else:
            assert funding_rate[enums.ExchangeConstantsFundingColumns.FUNDING_RATE.value].is_nan()
        if has_last_time:
            assert funding_rate[enums.ExchangeConstantsFundingColumns.LAST_FUNDING_TIME.value] <= self.get_time()
        else:
            assert funding_rate[enums.ExchangeConstantsFundingColumns.LAST_FUNDING_TIME.value] == constants.ZERO
        if has_next_rate:
            assert funding_rate[enums.ExchangeConstantsFundingColumns.PREDICTED_FUNDING_RATE.value]
        else:
            assert funding_rate[enums.ExchangeConstantsFundingColumns.PREDICTED_FUNDING_RATE.value].is_nan()
        if has_next_time:
            assert funding_rate[enums.ExchangeConstantsFundingColumns.NEXT_FUNDING_TIME.value] >= self.get_time()
        else:
            assert funding_rate[enums.ExchangeConstantsFundingColumns.NEXT_FUNDING_TIME.value] == constants.ZERO



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.orders cimport order_state
from octobot_trading.personal_data.orders.order_state cimport (
    OrderState,
)
from octobot_trading.personal_data.orders cimport order_group
from octobot_trading.personal_data.orders.order_group cimport (
    OrderGroup,
)
from octobot_trading.personal_data.orders cimport order
from octobot_trading.personal_data.orders.order cimport (
    Order,
)

from octobot_trading.personal_data.orders cimport orders_manager
from octobot_trading.personal_data.orders.orders_manager cimport (
    OrdersManager,
)

from octobot_trading.personal_data.orders cimport order_util
from octobot_trading.personal_data.orders cimport order_adapter
from octobot_trading.personal_data.orders cimport order_factory
from octobot_trading.personal_data.orders cimport types
from octobot_trading.personal_data.orders.types cimport (
    UnsupportedOrder,
    UnknownOrder,
    MarketOrder,
    SellMarketOrder,
    BuyMarketOrder,
    BuyLimitOrder,
    SellLimitOrder,
    LimitOrder,
    TakeProfitOrder,
    StopLossOrder,
    StopLossLimitOrder,
    TakeProfitLimitOrder,
    TrailingStopOrder,
    TrailingStopLimitOrder,
)

from octobot_trading.personal_data.orders cimport states
from octobot_trading.personal_data.orders.states cimport (
    CloseOrderState,
    CancelOrderState,
    OpenOrderState,
    PendingCreationOrderState,
    FillOrderState,
)

from octobot_trading.personal_data.orders cimport groups
from octobot_trading.personal_data.orders.groups cimport (
    BalancedTakeProfitAndStopOrderGroup,
    OneCancelsTheOtherOrderGroup,
)

from octobot_trading.personal_data.orders cimport channel
from octobot_trading.personal_data.orders.channel cimport (
    OrdersProducer,
    OrdersChannel,
    OrdersUpdater,
    OrdersUpdaterSimulator,
)

from octobot_trading.personal_data.orders.order_util cimport (
    is_valid,
    get_min_max_amounts,
    check_cost,
    total_fees_from_order_dict,
    get_max_order_quantity_for_price,
    get_fees_for_currency,
    parse_raw_fees,
    parse_order_status,
    parse_is_cancelled,
    get_pnl_transaction_source_from_order,
    is_stop_order,
    is_take_profit_order,
    is_associated_pending_order,
    get_order_quantity_currency,
    generate_order_id,
)
from octobot_trading.personal_data.orders.order_adapter cimport (
    adapt_price,
    adapt_quantity,
    adapt_order_quantity_because_quantity,
    adapt_order_quantity_because_price,
    split_orders,
    check_and_adapt_order_details_if_necessary,
    add_dusts_to_quantity_if_necessary,
)
from octobot_trading.personal_data.orders.order_factory cimport (
    create_order_from_raw,
    create_order_instance_from_raw,
    create_order_from_type,
    create_order_instance,
)

__all__ = [
    "Order",
    "is_valid",
    "get_min_max_amounts",
    "check_cost",
    "total_fees_from_order_dict",
    "get_max_order_quantity_for_price",
    "get_fees_for_currency",
    "parse_raw_fees",
    "parse_order_status",
    "parse_is_cancelled",
    "get_pnl_transaction_source_from_order",
    "is_stop_order",
    "is_take_profit_order",
    "is_associated_pending_order",
    "get_order_quantity_currency",
    "generate_order_id",
    "OrderState",
    "OrderGroup",
    "BalancedTakeProfitAndStopOrderGroup",
    "OneCancelsTheOtherOrderGroup",
    "OrdersUpdater",
    "adapt_price",
    "adapt_quantity",
    "adapt_order_quantity_because_quantity",
    "adapt_order_quantity_because_price",
    "split_orders",
    "check_and_adapt_order_details_if_necessary",
    "add_dusts_to_quantity_if_necessary",
    "create_order_from_raw",
    "create_order_instance_from_raw",
    "create_order_from_type",
    "create_order_instance",
    "OrdersProducer",
    "OrdersChannel",
    "OrdersManager",
    "OrdersUpdaterSimulator",
    "CloseOrderState",
    "CancelOrderState",
    "OpenOrderState",
    "PendingCreationOrderState",
    "FillOrderState",
    "UnsupportedOrder",
    "UnknownOrder",
    "MarketOrder",
    "SellMarketOrder",
    "BuyMarketOrder",
    "BuyLimitOrder",
    "SellLimitOrder",
    "LimitOrder",
    "TakeProfitOrder",
    "StopLossOrder",
    "StopLossLimitOrder",
    "TakeProfitLimitOrder",
    "TrailingStopOrder",
    "TrailingStopLimitOrder",
]



#  Drakkar-Software OctoBot-Private-Tentacles
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.constants as commons_constants
import octobot_trading.api as trading_api
import octobot_trading.exchanges as exchanges
import octobot_tentacles_manager.api as tentacles_manager_api


async def create_test_exchange_manager(
        config: object,
        exchange_name: str,
        is_spot_only: bool = True,
        is_margin: bool = False,
        is_future: bool = False,
        rest_only: bool = False,
        is_real: bool = True,
        is_sandboxed: bool = False,
        ignore_exchange_config: bool = True) -> exchanges.ExchangeManager:
    if ignore_exchange_config:
        # enable exchange name in config
        config[commons_constants.CONFIG_EXCHANGES][exchange_name] = {commons_constants.CONFIG_ENABLED_OPTION: True}

    builder = exchanges.create_exchange_builder_instance(config, exchange_name)
    builder.disable_trading_mode()
    builder.use_tentacles_setup_config(tentacles_manager_api.get_tentacles_setup_config())
    if is_spot_only:
        builder.is_spot_only()
    if is_margin:
        builder.is_margin()
    if is_future:
        builder.is_future()
    if rest_only:
        builder.is_rest_only()
    if is_sandboxed:
        builder.is_sandboxed(is_sandboxed)
    if is_real:
        builder.is_real()
    else:
        builder.is_simulated()
    await builder.build()
    return builder.exchange_manager


async def stop_test_exchange_manager(exchange_manager_instance: exchanges.ExchangeManager):
    trading_api.cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await asyncio_tools.wait_asyncio_next_cycle()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.enums as commons_enums
import octobot_commons.databases as commons_databases
import octobot_trading.enums as trading_enums


def set_plot_orders(ctx, value):
    ctx.plot_orders = value


async def disable_candles_plot(ctx, exchange_manager=None):
    storage = (exchange_manager or ctx.exchange_manager).storage_manager.candles_storage
    await storage.enable(False)
    await storage.clear_history()


async def store_orders(ctx, orders, exchange_manager,
                       chart=commons_enums.PlotCharts.MAIN_CHART.value,
                       x_multiplier=1000,
                       mode="markers",
                       kind="scattergl"):
    order_data = [
        {
            "x": order.creation_time * x_multiplier,
            "text": f"{order.order_type.name} {order.origin_quantity} {order.currency} at {order.origin_price}",
            "id": order.order_id,
            "symbol": order.symbol,
            "trading_mode": exchange_manager.trading_modes[0].get_name(),
            "type": order.order_type.name if order.order_type is not None else 'Unknown',
            "volume": float(order.origin_quantity),
            "y": float(order.created_last_price),
            "cost": float(order.total_cost),
            "state": order.state.state.value if order.state is not None else 'Unknown',
            "chart": chart,
            "kind": kind,
            "side": order.side.value,
            "mode": mode,
            "color": "red" if order.side is trading_enums.TradeOrderSide.SELL else "blue",
            "size": "10",
            "shape": "arrow-bar-left" if order.side is trading_enums.TradeOrderSide.SELL else "arrow-bar-right"
        }
        for order in orders
    ]
    await ctx.orders_writer.log_many(commons_enums.DBTables.ORDERS.value, order_data)


async def clear_orders_cache(writer):
    await _clear_table(writer, commons_enums.DBTables.ORDERS.value)
    await writer.clear()


async def clear_symbol_plot_cache(writer):
    await _clear_table(writer, commons_enums.DBTables.CACHE_SOURCE.value)
    await writer.clear()


async def _clear_table(writer, table, flush=True):
    await writer.delete(table, None)
    if flush:
        await writer.flush()


def get_shared_element(key):
    return commons_databases.GlobalSharedMemoryStorage.instance()[key]


def set_shared_element(key, element):
    commons_databases.GlobalSharedMemoryStorage.instance()[key] = element



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_trading.personal_data.trades.trade as trade_class
import octobot_trading.personal_data.orders.order_factory as order_factory
import octobot_trading.enums as enums
import octobot_trading.constants as constants


def create_trade_instance_from_raw(trader, raw_trade):
    try:
        order = order_factory.create_order_from_raw(trader, raw_trade)
        order.update_from_raw(raw_trade)
        if order.status is enums.OrderStatus.CANCELED:
            # ensure order is considered canceled
            order.consider_as_canceled()
        else:
            # ensure order is considered filled
            order.consider_as_filled()
        return create_trade_from_order(order)
    except KeyError:
        # Funding trade candidate
        return None


def create_trade_from_order(order,
                            close_status=None,
                            creation_time=0,
                            canceled_time=0,
                            executed_time=0):
    if close_status is not None:
        order.status = close_status
    trade = trade_class.Trade(order.trader)
    trade.update_from_order(order,
                            canceled_time=canceled_time,
                            creation_time=creation_time,
                            executed_time=executed_time)
    if trade.get_time() < constants.MINIMUM_VAL_TRADE_TIME:
        logging.get_logger("TradeFactory").error(f"Trade with invalid trade time ({trade.get_time()}) "
                                                 f"from order: {order}")
    return trade


def create_trade_instance(trader,
                          order_type,
                          symbol,
                          status=enums.OrderStatus.CLOSED,
                          order_id=None,
                          filled_price=constants.ZERO,
                          quantity_filled=constants.ZERO,
                          total_cost=constants.ZERO,
                          canceled_time=0,
                          creation_time=0,
                          executed_time=0):
    order = order_factory.create_order_from_type(trader=trader, order_type=order_type)
    order.update(order_type=order_type,
                 symbol=symbol,
                 current_price=filled_price,
                 quantity=quantity_filled,
                 price=filled_price,
                 order_id=trader.parse_order_id(order_id),
                 filled_price=filled_price,
                 quantity_filled=quantity_filled,
                 fee=None,  # TODO
                 total_cost=total_cost)
    return create_trade_from_order(order,
                                   close_status=status,
                                   canceled_time=canceled_time,
                                   creation_time=creation_time,
                                   executed_time=executed_time)


def create_trade_from_dict(trader, trade_dict):
    return trade_class.Trade.from_dict(trader, trade_dict)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import math
import pytest
import decimal

from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc
import octobot_trading.personal_data as personal_data
import octobot_trading.constants as constants
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_decimal_adapt_price():
    # will use symbol market
    symbol_market = {Ecmsc.PRECISION.value: {Ecmsc.PRECISION_PRICE.value: 4}}
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(0.0001))) == decimal.Decimal(
        str(0.0001))
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(0.00015))) == decimal.Decimal(
        str(0.0001))
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(0.005))) == decimal.Decimal(str(0.005))
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(1))) == decimal.Decimal(str(1))

    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(56.5128597145))) == decimal.Decimal(
        str(56.5128))
    assert personal_data.decimal_adapt_price(symbol_market,
                                             decimal.Decimal(str(1251.0000014576121234854513))) == decimal.Decimal(
        str(1251.0000))

    # will use default (CURRENCY_DEFAULT_MAX_PRICE_DIGITS)
    symbol_market = {Ecmsc.PRECISION.value: {}}
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(0.0001))) == decimal.Decimal(
        str(0.0001))
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(0.00015))) == decimal.Decimal(
        str(0.00015))
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(0.005))) == decimal.Decimal(str(0.005))
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(1))) == decimal.Decimal(
        str(1.0000000000000000000000001))
    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(1))) == decimal.Decimal(str(1))

    assert personal_data.decimal_adapt_price(symbol_market, decimal.Decimal(str(56.5128597145))) == decimal.Decimal(
        str(56.51285971))
    assert personal_data.decimal_adapt_price(symbol_market,
                                             decimal.Decimal(str(1251.0000014576121234854513))) == decimal.Decimal(
        str(1251.00000145))
    assert personal_data.decimal_adapt_price(symbol_market,
                                             decimal.Decimal(str(1251.0000014576121234854513)),
                                             True) == decimal.Decimal(
        str(1251.00000145))
    assert personal_data.decimal_adapt_price(symbol_market,
                                             decimal.Decimal(str(1251.0000014576121234854513)),
                                             False) == decimal.Decimal(
        str(1251.00000146))
    assert personal_data.decimal_adapt_price(symbol_market,
                                             decimal.Decimal(str(1251.000001451)),
                                             True) == decimal.Decimal(
        str(1251.00000145))
    assert personal_data.decimal_adapt_price(symbol_market,
                                             decimal.Decimal(str(1251.000001451)),
                                             False) == decimal.Decimal(
        str(1251.00000146))


async def test_decimal_check_and_adapt_order_details_if_necessary():
    symbol_market = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5,
                Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 1,
                Ecmsc.LIMITS_COST_MAX.value: 200
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.5,
                Ecmsc.LIMITS_PRICE_MAX.value: 50
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 8,
            Ecmsc.PRECISION_AMOUNT.value: 8
        }
    }

    invalid_cost_symbol_market = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.5,
                Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: None,
                Ecmsc.LIMITS_COST_MAX.value: None
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.5,
                Ecmsc.LIMITS_PRICE_MAX.value: 50
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 8,
            Ecmsc.PRECISION_AMOUNT.value: 8
        }
    }

    # correct min
    quantity = decimal.Decimal(str(0.5))
    price = decimal.Decimal(str(2))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # correct max
    quantity = decimal.Decimal(str(100))
    price = decimal.Decimal(str(2))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # correct
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(0.6))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # correct
    quantity = decimal.Decimal(str(3))
    price = decimal.Decimal(str(49.9))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # invalid price > but valid cost
    quantity = decimal.Decimal(str(1))
    price = decimal.Decimal(str(100))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # invalid price < but valid cost
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(0.1))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (quantity, price)]

    # invalid price > and invalid cost from exchange => use price => invalid
    quantity = decimal.Decimal(str(1))
    price = decimal.Decimal(str(100))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price,
                                                                            invalid_cost_symbol_market) == []

    # invalid price < and invalid cost from exchange => use price => invalid
    quantity = decimal.Decimal(str(1))
    price = decimal.Decimal(str(0.1))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price,
                                                                            invalid_cost_symbol_market) == []

    # invalid cost <
    quantity = decimal.Decimal(str(0.5))
    price = decimal.Decimal(str(1))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == []

    # invalid cost >
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(49))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (decimal.Decimal(str(1.83673469)), decimal.Decimal(str(49))),
        (decimal.Decimal(str(4.08163265)), decimal.Decimal(str(49))),
        (decimal.Decimal(str(4.08163265)), decimal.Decimal(str(49)))]

    # high cost but no max cost => valid
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(49))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (decimal.Decimal(str(1.83673469)), decimal.Decimal(str(49))),
        (decimal.Decimal(str(4.08163265)), decimal.Decimal(str(49))),
        (decimal.Decimal(str(4.08163265)), decimal.Decimal(str(49)))]

    # invalid cost with invalid price >=
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(50))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (decimal.Decimal(str(2)), decimal.Decimal(str(50))), (decimal.Decimal(str(4)), decimal.Decimal(str(50))),
        (decimal.Decimal(str(4)), decimal.Decimal(str(50)))]

    # invalid cost with invalid price > and invalid cost from exchange => use price => invalid
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(51))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price,
                                                                            invalid_cost_symbol_market) == []

    # invalid amount >
    quantity = decimal.Decimal(str(200))
    price = decimal.Decimal(str(5))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == [
        (decimal.Decimal(str(40.0)), decimal.Decimal(str(5))), (decimal.Decimal(str(40.0)), decimal.Decimal(str(5))),
        (decimal.Decimal(str(40.0)), decimal.Decimal(str(5))), (decimal.Decimal(str(40.0)), decimal.Decimal(str(5))),
        (decimal.Decimal(str(40.0)), decimal.Decimal(str(5)))]

    # invalid amount <
    quantity = decimal.Decimal(str(0.4))
    price = decimal.Decimal(str(5))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == []

    symbol_market = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.0000005,
                Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 0.00000001,
                Ecmsc.LIMITS_COST_MAX.value: 10
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.000005,
                Ecmsc.LIMITS_PRICE_MAX.value: 50
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 8,
            Ecmsc.PRECISION_AMOUNT.value: 8
        }
    }

    # correct quantity
    # to test _adapt_order_quantity_because_quantity
    quantity = decimal.Decimal(str(5000))
    price = decimal.Decimal(str(0.001))
    expected = [(decimal.Decimal(str(100.0)), decimal.Decimal(str(0.001)))] * 50
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == expected

    # price = 0 => no order
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(0))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price, symbol_market) == []

    symbol_market_without_max = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 0.0000005,
                Ecmsc.LIMITS_AMOUNT_MAX.value: None,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 0.00000001,
                Ecmsc.LIMITS_COST_MAX.value: None
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.000005,
                Ecmsc.LIMITS_PRICE_MAX.value: None
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 8,
            Ecmsc.PRECISION_AMOUNT.value: 8
        }
    }

    # high cost but no max cost => no split
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(49))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price,
                                                                            symbol_market_without_max) == [
               (decimal.Decimal(str(10)), decimal.Decimal(str(49)))]

    # high quantity but no max quantity => no split
    quantity = decimal.Decimal(str(10000000))
    price = decimal.Decimal(str(49))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price,
                                                                            symbol_market_without_max) == \
           [(decimal.Decimal(str(10000000)), decimal.Decimal(str(49)))]

    # high price but no max price => no split
    quantity = decimal.Decimal(str(10))
    price = decimal.Decimal(str(4900000))
    assert personal_data.decimal_check_and_adapt_order_details_if_necessary(quantity, price,
                                                                            symbol_market_without_max) == [
               (decimal.Decimal(str(10)), decimal.Decimal(str(4900000)))]


async def test_split_orders():
    symbol_market = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 1,
                Ecmsc.LIMITS_AMOUNT_MAX.value: 100,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 1,
                Ecmsc.LIMITS_COST_MAX.value: 30
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 0.001,
                Ecmsc.LIMITS_PRICE_MAX.value: 1000
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 1,
            Ecmsc.PRECISION_AMOUNT.value: 1
        }
    }

    symbol_market_without_max = {
        Ecmsc.LIMITS.value: {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: 1,
                Ecmsc.LIMITS_AMOUNT_MAX.value: None,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: 1,
                Ecmsc.LIMITS_COST_MAX.value: None
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: 1,
                Ecmsc.LIMITS_PRICE_MAX.value: None
            },
        },
        Ecmsc.PRECISION.value: {
            Ecmsc.PRECISION_PRICE.value: 1,
            Ecmsc.PRECISION_AMOUNT.value: 1
        }
    }
    max_cost = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_COST.value][Ecmsc.LIMITS_COST_MAX.value]
    max_quantity = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MAX.value]

    # normal situation, split because of cost
    total_price = decimal.Decimal(str(100))
    valid_quantity = decimal.Decimal(str(5))
    price = decimal.Decimal(str(20))
    assert personal_data.decimal_split_orders(total_price, max_cost, valid_quantity,
                                              max_quantity, price, valid_quantity, symbol_market) \
           == [(decimal.Decimal(str(0.5)), decimal.Decimal(str(20)))] + [
               (decimal.Decimal(str(1.5)), decimal.Decimal(str(20)))] * 3

    # normal situation, split because of quantity
    total_price = decimal.Decimal(str(5.0255))
    valid_quantity = decimal.Decimal(str(502.55))
    price = decimal.Decimal(str(0.01))
    assert personal_data.decimal_split_orders(total_price, max_cost, valid_quantity,
                                              max_quantity, price, valid_quantity, symbol_market) \
           == [(decimal.Decimal(str(2.5)), decimal.Decimal(str(0.01)))] + [
               (decimal.Decimal(str(100)), decimal.Decimal(str(0.01)))] * 5

    # missing info situation, split because of cost
    max_quantity = None
    total_price = decimal.Decimal(str(100))
    valid_quantity = decimal.Decimal(str(5))
    price = decimal.Decimal(str(20))
    assert personal_data.decimal_split_orders(total_price, max_cost, valid_quantity,
                                              max_quantity, price, valid_quantity, symbol_market_without_max) \
           == [(decimal.Decimal(str(0.5)), decimal.Decimal(str(20)))] + [
               (decimal.Decimal(str(1.5)), decimal.Decimal(str(20)))] * 3

    # missing info situation, split because of quantity
    max_quantity = symbol_market[Ecmsc.LIMITS.value][Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MAX.value]
    max_cost = None
    total_price = decimal.Decimal(str(5.0255))
    valid_quantity = decimal.Decimal(str(502.55))
    price = decimal.Decimal(str(0.01))
    assert personal_data.decimal_split_orders(total_price, max_cost, valid_quantity,
                                              max_quantity, price, valid_quantity, symbol_market_without_max) \
           == [(decimal.Decimal(str(2.5)), decimal.Decimal(str(0.01)))] + [
               (decimal.Decimal(str(100)), decimal.Decimal(str(0.01)))] * 5

    # missing info situation, can't split
    max_quantity = None
    max_cost = None
    total_price = decimal.Decimal(str(5.0255))
    valid_quantity = decimal.Decimal(str(502.55))
    price = decimal.Decimal(str(0.01))
    with pytest.raises(RuntimeError):
        assert personal_data.decimal_split_orders(total_price, max_cost, valid_quantity,
                                                  max_quantity, price, valid_quantity, symbol_market_without_max)


async def test_adapt_quantity():
    # will use symbol market
    symbol_market = {Ecmsc.PRECISION.value: {Ecmsc.PRECISION_AMOUNT.value: 4}}
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(0.0001))) == decimal.Decimal(
        str(0.0001))
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(0.00015))) == decimal.Decimal(
        str(0.0001))
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(0.005))) == decimal.Decimal(
        str(0.005))
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(1))) == decimal.Decimal(
        str(1.0000000000000000000000001))
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(1))) == decimal.Decimal(str(1))

    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(56.5128597145))) == decimal.Decimal(
        str(56.5128))
    assert personal_data.decimal_adapt_quantity(symbol_market,
                                                decimal.Decimal(str(1251.0000014576121234854513))) == decimal.Decimal(
        str(1251.0000))
    assert personal_data.decimal_adapt_quantity(symbol_market,
                                                decimal.Decimal(str(1251.0000014576121234854513)),
                                                True) == decimal.Decimal(
        str(1251.0000))
    assert personal_data.decimal_adapt_quantity(symbol_market,
                                                decimal.Decimal(str(1251.0000014576121234854513)),
                                                False) == decimal.Decimal(
        str(1251.0001))

    # will use default (0)
    symbol_market = {Ecmsc.PRECISION.value: {}}
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(0.0001))) == decimal.Decimal(str(0))
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(0.00015))) == decimal.Decimal(str(0))
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(0.005))) == decimal.Decimal(str(0))
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(1))) == decimal.Decimal(
        str(1.0000000000000000000000001))
    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(1))) == decimal.Decimal(str(1))

    assert personal_data.decimal_adapt_quantity(symbol_market, decimal.Decimal(str(56.5128597145))) == decimal.Decimal(
        str(56))
    assert personal_data.decimal_adapt_quantity(symbol_market,
                                                decimal.Decimal(str(1251.0000014576121234854513))) == decimal.Decimal(
        str(1251))


async def test_decimal_trunc_with_n_decimal_digits():
    assert personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(1.00000000001), 10) == decimal.Decimal(1)
    assert personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal("1.01"), 1, True) == decimal.Decimal(1)
    assert personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal("1.01"), 1, False) \
        == decimal.Decimal("1.1")
    assert float(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(1.00000000001), 10)) == 1
    assert float(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(1.00000000001), 11)) == 1.00000000001
    assert float(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(578.000145000156), 3)) == 578
    assert float(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(578.000145000156), 4)) == 578.0001
    assert float(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(578.000145000156), 7)) == 578.000145
    assert float(
        personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(11111111111111578.000145000156), 7)) == \
           11111111111111578.000145
    assert float(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(578.000145000156), 9)) == 578.000145
    assert float(
        personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(578.000145000156), 10)) == 578.0001450001
    # warning here, in python, float(578.000145000156) is not a finite number, therefore its decimal representation
    # is rounded to 578.000145000155. Use str in constructor to get the accurate representation (see next assert)
    assert float(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(578.000145000156), 12)) == \
           578.000145000155
    assert float(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal("578.000145000156"), 12)) == \
           578.000145000156
    assert float(
        personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(11111111111111578.000145000156), 12)) == \
           11111111111111578.000145000156
    assert personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal('0.00000000000001'), 7) == \
           constants.ZERO
    assert personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal('0.11111111111111'), 7) == \
           decimal.Decimal('0.1111111')

    assert math.isnan(personal_data.decimal_trunc_with_n_decimal_digits(decimal.Decimal(math.nan), 12))



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as trading_enums
import octobot_trading.constants as constants
import octobot_trading.personal_data as personal_data


_LOSING_ORDER_TYPES = [
    trading_enums.TradeOrderType.STOP_LOSS,
    trading_enums.TradeOrderType.STOP_LOSS_LIMIT,
]


def compute_win_rate(exchange_manager):
    lost_trades_count = constants.ZERO
    won_trades_count = constants.ZERO
    entries = constants.ZERO
    if exchange_manager.is_future:
        for trade in exchange_manager.exchange_personal_data.trades_manager.trades.values():
            if trade.status is trading_enums.OrderStatus.FILLED:
                if trade.reduce_only:
                    if trade.exchange_trade_type in _LOSING_ORDER_TYPES:
                        lost_trades_count += constants.ONE
                    else:
                        won_trades_count += constants.ONE
                else:
                    entries += constants.ONE
        total_exits = lost_trades_count + won_trades_count
        if total_exits > entries:
            # multiple take profits and SL not handled yet
            win_rate = constants.ZERO
        else:
            win_rate = won_trades_count / total_exits if total_exits else constants.ZERO
        if win_rate != constants.ZERO:
            return win_rate
        else:  # try fallback to transactions for trailing stops and multiple exits
            lost_trades_count = constants.ZERO
            won_trades_count = constants.ZERO
            for transaction in exchange_manager.exchange_personal_data.transactions_manager.transactions.values():
                if isinstance(transaction, personal_data.RealisedPnlTransaction) and transaction.is_closed_pnl():
                    if transaction.realised_pnl > constants.ZERO:
                        won_trades_count += constants.ONE
                    else:
                        lost_trades_count += constants.ONE
    else:
        for trade in exchange_manager.exchange_personal_data.trades_manager.trades.values():
            if trade.status is trading_enums.OrderStatus.FILLED and trade.side is trading_enums.TradeOrderSide.SELL:
                if trade.exchange_trade_type in _LOSING_ORDER_TYPES:
                    lost_trades_count += constants.ONE
                else:
                    won_trades_count += constants.ONE
    total_counted_trades = won_trades_count + lost_trades_count
    if total_counted_trades > constants.ZERO:
        return won_trades_count / total_counted_trades
    return constants.ZERO



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.orders.types cimport limit
from octobot_trading.personal_data.orders.types.limit cimport (
    BuyLimitOrder,
    SellLimitOrder,
    LimitOrder,
    TakeProfitOrder,
    StopLossOrder,
    StopLossLimitOrder,
    TakeProfitLimitOrder,
)
from octobot_trading.personal_data.orders.types cimport trailing
from octobot_trading.personal_data.orders.types.trailing cimport (
    TrailingStopOrder,
    TrailingStopLimitOrder,
)
from octobot_trading.personal_data.orders.types cimport market
from octobot_trading.personal_data.orders.types.market cimport (
    MarketOrder,
    SellMarketOrder,
    BuyMarketOrder,
)
from octobot_trading.personal_data.orders.types cimport unknown_order
from octobot_trading.personal_data.orders.types.unknown_order cimport (
    UnknownOrder,
)
from octobot_trading.personal_data.orders.types cimport unsupported_order
from octobot_trading.personal_data.orders.types.unsupported_order cimport (
    UnsupportedOrder,
)

__all__ = [
    "UnsupportedOrder",
    "UnknownOrder",
    "MarketOrder",
    "SellMarketOrder",
    "BuyMarketOrder",
    "BuyLimitOrder",
    "SellLimitOrder",
    "LimitOrder",
    "TakeProfitOrder",
    "StopLossOrder",
    "StopLossLimitOrder",
    "TakeProfitLimitOrder",
    "TrailingStopOrder",
    "TrailingStopLimitOrder",
]

